#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <time.h>

#define NITERS 1000000000

// Wall clock
int read_clock(struct timeval *t)
{
    return gettimeofday(t, NULL);
}

int elapsed_time(struct timeval *start, struct timeval *end)
{
    return (int)(end->tv_sec  - start->tv_sec) * 1e6 + 
		   (int)(end->tv_usec - start->tv_usec);
}

int main( int argc, char *argv[] )
{ 
    struct timeval t0, t1;
    
    int i;
    double eps = 1 + 1e-8;

    // Parse arguments (TEST_CASE in range [0,4])
    if ( argc != 2 )
    {
        fprintf( stderr, "Usage: %s TEST_CASE (0..4)\n", argv[0] );
        exit( -1 );
    }

    int TEST_CASE = atoi( argv[1] );
    if ( TEST_CASE < 0 || TEST_CASE > 4 )
    {
        fprintf( stderr, "Usage: %s TEST_CASE (0..4)\n", argv[0] );
        exit( -1 );
    }

    // Set up for multiple test cases:
    //   cycle_len: number of independent operations (nops)
    //   cycle_mask: bit_mask to cycle from 0 to nops
    int cycle_len[5]  = {    1,    2,    4,    8,   16};
    int cycle_mask[5] = { 0x00, 0x01, 0x03, 0x07, 0x0f};

    // Based on test case, select corresponding cycle_len and cycle_mask
    int nops = cycle_len[ TEST_CASE ];
    int mask = cycle_mask[ TEST_CASE ];

    // Array to store the result of the n independent ops
    double A[ nops ];
    for ( i = 0; i < nops; i++ )
        A[i] = 1.0;

    // Run and time the actual computation
    read_clock( &t0 );
    int cycle_idx = 0;
    for( i = 0; i < NITERS; i++ )
    {
        A[ cycle_idx ] = A[ cycle_idx ] * eps;
        cycle_idx = (cycle_idx + 1) & mask;
    }
    // Reduction
    double red = 1.0;
    for ( i = 0; i < nops; i++ )
        red *= A[i];
    read_clock( &t1 );

    // Do something with the result
    fprintf(stderr, "Total:%.5e\n", red);

    printf(" [%2d] independent operations\n", nops);
    printf(" Time: %.3f seconds\n", elapsed_time( &t0, &t1 ) * 1e-6 );

    return 0;
}
