#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <time.h>

#include <immintrin.h>

#define NITERS 1000000000
#define VLEN 4

// Wall clock
int read_clock(struct timeval *t)
{
    return gettimeofday(t, NULL);
}

int elapsed_time(struct timeval *start, struct timeval *end)
{
    return (int)(end->tv_sec  - start->tv_sec) * 1e6 + 
		   (int)(end->tv_usec - start->tv_usec);
}

int main( int argc, char *argv[] )
{ 
    struct timeval t0, t1;
    
    int i;
    double eps = 1 + 1e-8;
    __m256d eps_v = _mm256_set_pd( eps, eps, eps, eps );
    
    // Parse arguments (TEST_CASE in range [0,4])
    if ( argc != 2 )
    {
        fprintf( stderr, "Usage: %s TEST_CASE\n", argv[0] );
        exit( -1 );
    }
    int TEST_CASE = atoi( argv[1] );

    // Set up for multiple test cases:
    //   cycle_len: number of independent operations (nops)
    //   cycle_mask: bit_mask to cycle from 0 to nops
    int cycle_len[5]  = {    1,    2,    4,    8,   16};
    int cycle_mask[5] = { 0x00, 0x01, 0x03, 0x07, 0x0f};

    // Based on test case, select corresponding cycle_len and cycle_mask
    int nops = cycle_len[ TEST_CASE ];
    int mask = cycle_mask[ TEST_CASE ];

    // Array to store the result of the n independent ops
    // double A[ nops ];
    // for ( i = 0; i < nops; i++ )
    //    A[i] = 0.0;
    __m256d A[ nops ];
    for ( i = 0; i < nops; i++ )
       A[i] = _mm256_set_pd( 1.0, 1.0, 1.0, 1.0 );

    // Run and time the actual computation
    read_clock( &t0 );
    int cycle_idx = 0;
    // for( i = 0; i < NITERS; i++ )
    for( i = 0; i < NITERS/VLEN; i++ )
    {
        // A[ cycle_idx ] = A[ cycle_idx ] * eps;
        A[ cycle_idx ] = _mm256_mul_pd( A[ cycle_idx ], eps_v );
        cycle_idx = (cycle_idx + 1) & mask;
    }
    // Reduction
    //
    // double red = 0.0;
    // for ( i = 0; i < nops; i++ )
    //     red += A[i];
    // fprintf(stderr, "Total:%.16e\n", red);
    __m256d red_v = _mm256_set_pd( 1.0, 1.0, 1.0, 1.0 );
    for ( i = 0; i < nops; i++ )
        red_v = _mm256_mul_pd( red_v, A[i] );
    double red = 1.0;
    for ( i = 0; i < VLEN; i++ )
        red = red * ((double*)&red_v)[i];
    read_clock( &t1 );

    // Do something with the result
    fprintf(stderr, "Total:%.5e\n", red);

    printf(" [%2d] independent operations\n", nops);
    printf(" Time: %.3f seconds\n", elapsed_time( &t0, &t1 ) * 1e-6 );

    return 0;
}
